#!/usr/bin/env python

import os
import sys
import time
import shlex
import signal
import argparse
import tempfile
import traceback
import subprocess as sp
from datetime import datetime

"""
x265ify.py: transcode video files to h265, third edition
2020-09-30, v1.0, ed <irc.rizon.net>, MIT-Licensed
https://ocv.me/dev/?x265ify.py

this replaces https://ocv.me/dev/?x265ify.bat

existing features carried over from the batch ver:
- CTRL-C once to terminate an encode cleanly

new features since the batch ver:
- full support for filenames with unicode and special characters
- runs FFmpeg at minimal priority
- support for x264 and vp9, tho x265 is generally still best

ctrl-f "cfg:" to change defaults
"""

try:
    import ctypes
    from ctypes import wintypes

    _GetLongPathNameW = ctypes.windll.kernel32.GetLongPathNameW
    _GetLongPathNameW.argtypes = [wintypes.LPCWSTR, wintypes.LPWSTR, wintypes.DWORD]
    _GetLongPathNameW.restype = wintypes.DWORD
    
    WINDOWS = True
except:
    WINDOWS = False


CLEANUP = []


def longify(sPath):
    sz = 0
    for _ in range(64):
        buf = ctypes.create_unicode_buffer(sz)
        sz2 = _GetLongPathNameW(sPath, buf, sz)
        if sz2 == 0:
            return None
        if sz >= sz2:
            return buf.value
        
        sz = sz2

    raise Exception("uhh")


def need_shortnames():
    print("\nthis script requires the python drophandler to provide 8.3 names, so please switch from the WSH to the EXE handler (because WSH doesn't support unicode)")
    reg_path = os.path.abspath(sys.argv[0]) + ".reg"
    try:
        with open(reg_path, 'wb') as f:
            f.write(b"\xff\xfe" + r"""Windows Registry Editor Version 5.00

[HKEY_CLASSES_ROOT\Python.File\shellex\DropHandler]
@="{86C86720-42A0-1069-A2E8-08002B30309D}"

[HKEY_LOCAL_MACHINE\SOFTWARE\Classes\Python.File\shellex\DropHandler]
@="{86C86720-42A0-1069-A2E8-08002B30309D}"

; orig: wsh {60254CA5-953B-11CF-8C96-00AA00B8708C}
; works: exe {86C86720-42A0-1069-A2E8-08002B30309D}
""".encode('utf-16le'))

        print(f"\nthere's a registry fix you can apply here:\n    {reg_path}\n")
        print(f"hit ENTER to apply the fix, or CTRL-C to abort\n")
        input()
        sp.check_call(['regedit', reg_path], shell=True)
        print(f"\nok now you have to restart explorer.exe for that to take effect,\nor alternatively reassign the open-with for .py files\n")
        print(f"(hit ENTER to close this window)\n")

    except:
        print(f'\ntried (and failed) to create a registry fix at this path:\n    {reg_path}\n')


def sighandler(sig, frame):
    pass


def transcode(ar, fc_path, src, name_extra):
    preset = ar.p
    crf = ar.c
    ia = ar.ia
    oa = ar.oa
    vf = ar.vf

    # cfg: seek 9 seconds into video if "ID INVADED" in filename
    if 'ID INVADED' in src:
        ia += ' -ss 9'

    if 'x264' in preset:
        enc = 'h264'
        preset = 'veryslow'
    elif 'vp9' in preset:
        enc, preset, maxrate = preset.split('-')
        spd = preset.lstrip('s')
    else:
        enc = 'h265'

    # cfg: write output to ~/enc
    outdir = os.path.join(os.path.expanduser('~'), 'enc')

    outfile = os.path.join(outdir, os.path.basename(src))
    #outfile += f".h265-720p-{preset}{crf}.mkv"
    outfile += f".{enc}-{name_extra}{preset}-{crf}.mkv"
    
    logfile = outfile + ".log"
    tmp_log = f"x265ify-{int(time.time()*1000)/1000.}-{os.getpid()}.log"
    tmp_log = os.path.join(os.path.expanduser('~'), tmp_log)
    ff_tmplog = tmp_log.replace('\\', '\\\\').replace(':', '\\:')

    if os.path.exists(outfile):
        return False

    lockdir = outfile + ".lock"
    try:
        os.mkdir(lockdir)
        CLEANUP.append(lockdir)
    except:
        return False

    try:
        os.unlink(logfile)
    except:
        pass

    env = os.environ.copy()
    env["FONTCONFIG_FILE"] = fc_path
    env["FFREPORT"] = f"level=32:file={ff_tmplog}"

    src_sz = os.path.getsize(src)
    src_mod = os.path.getmtime(src)
    src_mod = datetime.utcfromtimestamp(src_mod)
    src_mod = src_mod.strftime("%Y-%m-%d %H:%M:%S")

    print(f"processing [{src}]")
    print(f"writing to [{outfile}]")
    print(f"logfile at [{tmp_log}]")
    print()

    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-vsync", "vfr"
    ]
    cmd.extend(shlex.split(ia))
    cmd.extend([
        "-i", src,
        "-map", "0",
        "-c", "copy",
        "-vf", vf
    ])

    if enc == 'h265':
        cmd.extend([
            "-c:v", "libx265",
            "-preset", preset
        ])
    elif enc == 'x264':
        cmd.extend([
            "-c:v", "libx264",
            "-preset", preset
        ])
    else:
        cmd.extend([
            "-c:v", "libvpx-vp9",
            "-b:v", maxrate,
            "-speed", spd,
            "-auto-alt-ref", "1",
            "-lag-in-frames", "25",
            "-frame-parallel", "0",
            "-row-mt", "1"
        ])

    cmd.extend([
        "-crf", str(crf),
    ])
    cmd.extend(shlex.split(oa))
    cmd.extend([
        "-pix_fmt", "yuv420p",
        "-max_muxing_queue_size", "99999",
        outfile
    ])

    p = sp.Popen(cmd, env=env)
    p.wait()
    if p.returncode == 0:
        status = 'OK'
    else:
        status = f'ERROR({p.returncode})'

    footer = f"\nx265ify.py {status}: original file last-modified {src_mod}, size {src_sz}\n"
    print(footer)

    os.rename(tmp_log, logfile)
    with open(logfile, 'ab+') as f:
        f.write(footer.encode('utf-8'))

    os.rmdir(lockdir)
    CLEANUP.remove(lockdir)
    return True


def cli_main():
    if len(sys.argv) < 2:
        print('drop one or more video files onto this python-file,')
        print('or see -h for help')
        sys.exit(1)

    # cfg: output quality; last value wins,
    #   anything above 20 has visible quality loss,
    #   48 is probably the highest you'll want for x265 (tiny files)
    #   52 is probably the highest you'll want for vp9
    _crf = [
        20,
        28,
        34,
        48,
        52,
        60,
        36,
    ]

    # cfg: time to spend, more time = better quality and smaller files,
    #   possible values from fastest to slowest:
    #   ultrafast superfast veryfast faster fast medium slow slower veryslow placebo
    # if vp9, format is vp9-s$speed-$max_bitrate
    # if x264, format is x264
    _preset = [
        "x264",
        "vp9-s1-3000k",
        "medium",
        "veryslow",
        "slow",
        "slower",
    ]

    # cfg: extra arguments on the input-side,
    #   -threads 4 lets FFmpeg hse 4 cores "generally"
    _inflags = [
        "-threads 4"
    ]

    # cfg: extra arguments on the output-side,
    # -r 30 sets output fps = 30,
    # x265 "ignores" -threads, pools=4 allocates 4 cores on one numa,
    #   https://x265.readthedocs.io/en/latest/cli.html#performance-options
    #   https://trac.ffmpeg.org/ticket/3730#comment:10
    # x265 -tune animation adds a lot of ringing
    _outflags = [
        "-tune animation -x265-params pools=4:vbv-maxrate=4000",
        "-tune animation -x265-params pools=4:vbv-maxrate=6000",
        "-x265-params pools=4:vbv-maxrate=7000:vbv-bufsize=14000",
        "-x265-params pools=4:vbv-maxrate=3000:vbv-bufsize=6000 -r 30",
        "-x265-params vbv-maxrate=3000:vbv-bufsize=6000",
        "-x265-params pools=4:vbv-maxrate=3000:vbv-bufsize=6000",
    ]

    # cfg: video filters to apply,
    #   mpdecimate is really good for mixed-fps sources (online music parties)
    _filters = [
        "pp7=3,vaguedenoiser=4",
        "pp7=3",
        "pp7=3,fftdnoiz=10:prev=1:next=1",
        "mpdecimate=frac=0.01,pp7=1,minterpolate=30000/1001:mci:mc_mode=aobmc:me_mode=bidir:vsbmc=1:mb_size=10",
        "copy",
        "pp=hb/vb/dr/al",
        "pp=ac,deflicker",
        "pp=ac,pp7=3",
        "pp=ac",
        "mpdecimate=frac=0.01,pp=ac,pp7=1",
    ]

    # cfg: resize wideo to WIDTH pixels 
    _width = [
        1280,
        None,
    ]

    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    ap.add_argument("-x", action="store_true", help="immediate exit (cli-mode)")
    ap.add_argument("-c", metavar="CRF", type=int, default=_crf[-1], help="quality level (higher=worse)")
    ap.add_argument("-p", metavar="PRESET", default=_preset[-1], help="preset")
    ap.add_argument("-ia", metavar="ARGS", default=_inflags[-1], help="input args")
    ap.add_argument("-oa", metavar="ARGS", default=_outflags[-1], help="output args")
    ap.add_argument("-vf", metavar="FILTERS", default=_filters[-1], help="video filters")
    ap.add_argument("-sz", metavar="RES", default=_width[-1], help="resize to NNN:-16")
    ap.add_argument("files", metavar="FILEPATH", nargs="+", help="files to encode")
    ar = ap.parse_args()

    name_extra = ''
    if ar.sz:
        ar.vf += f",scale={ar.sz}:-16"
        name_extra += f"{ar.sz}x-"
    
    # hardsub support
    fc_path = os.path.join(tempfile.gettempdir(), 'ffhs-fonts.conf')
    with open(fc_path, 'wb') as f:
        f.write(r"""<?xml version="1.0"?>
<!DOCTYPE fontconfig SYSTEM "fonts.dtd">
<fontconfig><dir>.</dir></fontconfig>
""".encode('utf-8'))
    
    # lower piroity; will inherit  ['idle', 'below normal']
    if WINDOWS:
        sp.check_call(['wmic', 'process', 'where', f'processid={os.getpid()}', 'call', 'setpriority', 'idle'])
    else:
        os.nice(15)

    # ignore ctrl-c
    signal.signal(signal.SIGINT, sighandler)

    files = []
    for n, short_src in enumerate(sys.argv[1:]):
        if WINDOWS:
            src = longify(short_src)
            print(f"got file #{n}: {short_src}\n  {src}\n")
        else:
            src = short_src
            print(f"got file #{n}: {src}")

        if src is None or not os.path.lexists(src):
            if WINDOWS:
                need_shortnames()
            else:
                print(f"could not open file [{src}]")
            
            sys.exit(1)

        if not os.path.isfile(src):  # follows symlinks
            print("only files please, no folders")
            sys.exit(1)

        files.append(src)

    for src in files:
        try:
            if not transcode(ar, fc_path, src, name_extra):
                print("*** file skipped ***")
        except:
            raise
        finally:
            try:
                os.rmdir(CLEANUP[0])
            except:
                pass


def gui_main():
    try:
        cli_main()
        sys.exit(0)
        input()
    except SystemExit as ex:
        if ex.code == 0:
            print('\n(finished, press ENTER to close)')
        else:
            print(f'\n(error {ex.code}, press ENTER to close)')
        
        input()
        raise
    except KeyboardInterrupt:
        pass
    except:
        traceback.print_exc()
        print('\n(crash, press ENTER to close)')
        input()
        sys.exit(1)


def main():
    for arg in sys.argv[1:]:
        if arg.startswith('-') and 'x' in arg:
            return cli_main()
    
    gui_main()


if __name__ == '__main__':
    main()


"""
# move all 60fps videos into a subfolder named "60fps":
mkdir 60fps; for f in *.webm *.mkv; do printf '%s\n' "$f"; ffmpeg -i "$f" 2>&1 | grep ' fps, '; done | tee log | grep ', 60 fps, ' -B1 | grep -vE '^ +Stream #|^--$' | while IFS= read -r x; do mv -- "$x" 60fps/; done
"""

